/*
  Copyright (C) 2009, 2010 and 2013 and 2015 Chris Vine

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <stdlib.h>
#include <string.h>

#include <vector>
#include <fstream>
#include <ios>
#include <ostream>

#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h> // the key codes are here

#include <glib.h>

#include "mount_entries.h"
#include "mount_entries_icons.h"
#include "dialogs.h"
#include "utils/tree_path_handle.h"

#include <c++-gtk-utils/shared_handle.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif


const int standard_size = 24;

namespace {
namespace ModelColumns {
  enum {device, label, cols_num};
}
} // anonymous namespace

void MountEntriesCB::mount_entries_button_clicked(GtkWidget* widget_p, void* data) {
  MountEntries* instance_p = static_cast<MountEntries*>(data);
  
  if (widget_p == instance_p->ok_button_p) {
    if (instance_p->changed) instance_p->save_list();
    instance_p->close();
  }
  else if (widget_p == instance_p->cancel_button_p) {
    instance_p->changed = false;
    instance_p->close();
  }
  else if (widget_p == instance_p->add_button_p) {
    instance_p->add_entry_prompt();
  }
  else if (widget_p == instance_p->modify_button_p) {
    instance_p->modify_entry_prompt();
  }
  else if (widget_p == instance_p->delete_button_p) {
    instance_p->delete_entry_prompt();
  }
  else if (widget_p == instance_p->up_button_p) {
    instance_p->move_up_impl();
  }
  else if (widget_p == instance_p->down_button_p) {
    instance_p->move_down_impl();
  }
  else {
    write_error("Callback error in MountEntriesCB::mount_entries_button_clicked()\n", false);
    instance_p->changed = false;
    instance_p->close();
  }
}

void MountEntriesCB::mount_entries_drag_n_drop(GtkTreeModel*, GtkTreePath*, void* data) {
  static_cast<MountEntries*>(data)->changed = true;
}

MountEntries::MountEntries(GtkWindow* parent_p): WinBase(gettext("Mount-gtk: mount table"),
							 prog_config.window_icon_h,
 							 true, parent_p),
						 changed(false) {

  GtkTable* table_p = GTK_TABLE(gtk_table_new(2, 1, false));
  gtk_container_add(GTK_CONTAINER(get_win()), GTK_WIDGET(table_p));

  GtkTable* list_table_p = GTK_TABLE(gtk_table_new(6, 2, false));
  gtk_table_attach(table_p, GTK_WIDGET(list_table_p),
		   0, 1, 0, 1,
		   GtkAttachOptions(GTK_FILL | GTK_EXPAND), GtkAttachOptions(GTK_FILL | GTK_EXPAND),
		   standard_size/3, standard_size/3);

  GtkWidget* button_box_p = gtk_hbutton_box_new();
  gtk_table_attach(table_p, button_box_p,
		   0, 1, 1, 2,
		   GtkAttachOptions(GTK_FILL | GTK_EXPAND), GTK_SHRINK,
		   standard_size/3, standard_size/3);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box_p), GTK_BUTTONBOX_END);
  gtk_box_set_spacing(GTK_BOX(button_box_p), standard_size/2);

  cancel_button_p = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_container_add(GTK_CONTAINER(button_box_p), cancel_button_p);
  ok_button_p = gtk_button_new_from_stock(GTK_STOCK_OK);
  gtk_container_add(GTK_CONTAINER(button_box_p), ok_button_p);

  add_button_p = gtk_button_new();
  gtk_table_attach(list_table_p, add_button_p,
		   0, 1, 0, 1,
		   GTK_SHRINK, GTK_SHRINK,
		   standard_size/2, 0);
  modify_button_p = gtk_button_new();
  gtk_table_attach(list_table_p, modify_button_p,
		   0, 1, 1, 2,
		   GTK_SHRINK, GTK_SHRINK,
		   standard_size/2, 0);
  delete_button_p = gtk_button_new();
  gtk_table_attach(list_table_p, delete_button_p,
		   0, 1, 2, 3,
		   GTK_SHRINK, GTK_SHRINK,
		   standard_size/2, 0);
  up_button_p = gtk_button_new();
  gtk_table_attach(list_table_p, up_button_p,
		   0, 1, 3, 4,
		   GTK_SHRINK, GTK_SHRINK,
		   standard_size/2, 0);
  down_button_p = gtk_button_new();
  gtk_table_attach(list_table_p, down_button_p,
		   0, 1, 4, 5,
		   GTK_SHRINK, GTK_SHRINK,
		   standard_size/2, 0);
  GtkWidget* dummy_p = gtk_label_new(0);
  gtk_table_attach(list_table_p, dummy_p,
		   0, 1, 5, 6,
		   GTK_SHRINK, GTK_EXPAND,
		   standard_size/2, 0);
  GtkScrolledWindow* scrolled_window_p = GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(0, 0));
  gtk_table_attach(list_table_p, GTK_WIDGET(scrolled_window_p),
		   1, 2, 0, 6,
		   GtkAttachOptions(GTK_FILL | GTK_EXPAND), GtkAttachOptions(GTK_FILL | GTK_EXPAND),
		   0, 0);

  gtk_scrolled_window_set_shadow_type(scrolled_window_p, GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(scrolled_window_p, GTK_POLICY_ALWAYS,
				 GTK_POLICY_ALWAYS);

  list_store_h =
    GobjHandle<GtkTreeModel>(GTK_TREE_MODEL(gtk_list_store_new(ModelColumns::cols_num,
							       G_TYPE_STRING, G_TYPE_STRING)));
  // populate the model (containing the entries list)
  read_list();

  // create the tree view
  tree_view_p = GTK_TREE_VIEW(gtk_tree_view_new_with_model(list_store_h));
  gtk_container_add(GTK_CONTAINER(scrolled_window_p), GTK_WIDGET(tree_view_p));

  // provide renderers for tree view, pack into tree view columns
  // and connect to the tree model columns
  GtkCellRenderer* renderer_p = gtk_cell_renderer_text_new();
  GtkTreeViewColumn* column_p =
    gtk_tree_view_column_new_with_attributes(gettext("Device"), renderer_p,
					     "text", ModelColumns::device,
					     static_cast<void*>(0));
  gtk_tree_view_append_column(tree_view_p, column_p);
  
  renderer_p = gtk_cell_renderer_text_new();
  column_p = gtk_tree_view_column_new_with_attributes(gettext("Label"), renderer_p,
						      "text", ModelColumns::label,
						      static_cast<void*>(0));
  gtk_tree_view_append_column(tree_view_p, column_p);

  GtkTreeSelection* selection_p = gtk_tree_view_get_selection(tree_view_p);
  // single line selection
  gtk_tree_selection_set_mode(selection_p, GTK_SELECTION_SINGLE);

  // make drag and drop work
  gtk_tree_view_set_reorderable(tree_view_p, true);

  // have any drag and drop saved to the set-up file
  g_signal_connect(G_OBJECT(list_store_h.get()), "row_deleted",
		   G_CALLBACK(MountEntriesCB::mount_entries_drag_n_drop), this);

  // bring up the icon size registered in MainWindow::MainWindow()
  GtkIconSize mount_gtk_button_size = gtk_icon_size_from_name("MOUNT_GTK_BUTTON_SIZE");

  GtkWidget* image_p;
  image_p = gtk_image_new_from_stock(GTK_STOCK_ADD, mount_gtk_button_size);
  gtk_container_add(GTK_CONTAINER(add_button_p), image_p);

  { // provide a scope block for the GobjHandle
    GobjHandle<GdkPixbuf> pixbuf_h(gdk_pixbuf_new_from_xpm_data(modify_xpm));
    image_p = gtk_image_new_from_pixbuf(pixbuf_h);
    gtk_container_add(GTK_CONTAINER(modify_button_p), image_p);
  }

  image_p = gtk_image_new_from_stock(GTK_STOCK_DELETE, mount_gtk_button_size);
  gtk_container_add(GTK_CONTAINER(delete_button_p), image_p);

  image_p = gtk_image_new_from_stock(GTK_STOCK_GO_UP, mount_gtk_button_size);
  gtk_container_add(GTK_CONTAINER(up_button_p), image_p);

  image_p = gtk_image_new_from_stock(GTK_STOCK_GO_DOWN, mount_gtk_button_size);
  gtk_container_add(GTK_CONTAINER(down_button_p), image_p);

  gtk_widget_set_tooltip_text(add_button_p, gettext("Add new mount item"));
  gtk_widget_set_tooltip_text(modify_button_p, gettext("Modify mount item"));
  gtk_widget_set_tooltip_text(delete_button_p, gettext("Delete mount item"));
  gtk_widget_set_tooltip_text(up_button_p, gettext("Move mount item up"));
  gtk_widget_set_tooltip_text(down_button_p, gettext("Move mount item down"));

  g_signal_connect(G_OBJECT(ok_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  g_signal_connect(G_OBJECT(cancel_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  g_signal_connect(G_OBJECT(add_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  g_signal_connect(G_OBJECT(modify_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  g_signal_connect(G_OBJECT(delete_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  g_signal_connect(G_OBJECT(up_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  g_signal_connect(G_OBJECT(down_button_p), "clicked",
		   G_CALLBACK(MountEntriesCB::mount_entries_button_clicked), this);
  
#if GTK_CHECK_VERSION(2,20,0)
  gtk_widget_set_can_default(ok_button_p, true);
  gtk_widget_set_can_default(cancel_button_p, true);

  gtk_widget_set_can_default(add_button_p, false);
  gtk_widget_set_can_default(delete_button_p, false);
  gtk_widget_set_can_focus(up_button_p, false);
  gtk_widget_set_can_focus(down_button_p, false);
#else
  GTK_WIDGET_SET_FLAGS(ok_button_p, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS(cancel_button_p, GTK_CAN_DEFAULT);

  GTK_WIDGET_UNSET_FLAGS(add_button_p, GTK_CAN_DEFAULT);
  GTK_WIDGET_UNSET_FLAGS(delete_button_p, GTK_CAN_DEFAULT);
  GTK_WIDGET_UNSET_FLAGS(up_button_p, GTK_CAN_FOCUS);
  GTK_WIDGET_UNSET_FLAGS(down_button_p, GTK_CAN_FOCUS);
#endif

  gtk_container_set_border_width(GTK_CONTAINER(table_p), standard_size/3);
  gtk_window_set_type_hint(get_win(), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_window_set_position(get_win(), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_widget_grab_focus(cancel_button_p);
  gtk_window_set_default_size(get_win(), standard_size * 15, standard_size * 8);
  gtk_widget_show_all(GTK_WIDGET(get_win()));
}

int MountEntries::get_exec_val(void) const {
  return changed;
}

void MountEntries::on_delete_event(void) {
  changed = false;
  close();
}

void MountEntries::add_entry_prompt(void) {

  TableEntry dialog(get_win(), 0, 0);
  if (dialog.exec()) add_entry(dialog);
}

void MountEntries::add_entry(const TableEntry& dialog) {

  // this comes pre-stripped
  std::pair<std::string, std::string> result = dialog.get_result();

  if (!result.first.empty()) { // this test should be unnecessary, but ...

    GtkTreeIter iter;
    gtk_list_store_append(GTK_LIST_STORE(list_store_h.get()), &iter);
    gtk_list_store_set(GTK_LIST_STORE(list_store_h.get()), &iter,
		       ModelColumns::device, result.first.c_str(),
		       ModelColumns::label, result.second.c_str(),
		       -1);
    changed = true;
  }
}

void MountEntries::modify_entry_prompt() {
  GtkTreeIter iter;
  GtkTreeModel* model_p;
  GtkTreeSelection* selection_p = gtk_tree_view_get_selection(tree_view_p);
  gchar* device_p = 0;
  gchar* label_p = 0;

  if (gtk_tree_selection_get_selected(selection_p, &model_p, &iter)) {
    gtk_tree_model_get(model_p, &iter,
		       ModelColumns::device, &device_p,
		       ModelColumns::label, &label_p,
		       -1);
    if (device_p && label_p) {
      try {
	TableEntry dialog(get_win(), device_p, label_p);
	if (dialog.exec()) modify_entry(dialog);
      }
      catch (...) {
	g_free(device_p);
	g_free(label_p);
	throw;
      }
    }
    g_free(device_p);
    g_free(label_p);
  }
  else beep();
}

void MountEntries::modify_entry(const TableEntry& dialog) {

  // this comes pre-stripped
  std::pair<std::string, std::string> result = dialog.get_result();

  if (!result.first.empty()) { // this test should be unnecessary, but ...

    GtkTreeIter iter;
    GtkTreeSelection* selection_p = gtk_tree_view_get_selection(tree_view_p);

    if (gtk_tree_selection_get_selected(selection_p, 0, &iter)) {
      gtk_list_store_set(GTK_LIST_STORE(list_store_h.get()), &iter,
			 ModelColumns::device, result.first.c_str(),
			 ModelColumns::label, result.second.c_str(),
			 -1);
      changed = true;
    }
  }
}

void MountEntries::delete_entry_prompt(void) {
  GtkTreeIter iter;
  GtkTreeSelection* selection_p(gtk_tree_view_get_selection(tree_view_p));

  if (gtk_tree_selection_get_selected(selection_p, 0, &iter)) {
    PromptDialog dialog(gettext("Delete selected mount entry?"),
			gettext("Gtk-mount: delete entry"),
			get_win());
    if (dialog.exec()) delete_entry();
  }
  else beep();
}

void MountEntries::delete_entry(void) {

  GtkTreeIter iter;
  GtkTreeModel* model_p;
  GtkTreeSelection* selection_p(gtk_tree_view_get_selection(tree_view_p));

  if (gtk_tree_selection_get_selected(selection_p, &model_p, &iter)) {
    // delete the address by removing it from the list store
    gtk_list_store_remove(GTK_LIST_STORE(model_p), &iter);
    changed = true;
  }
}

void MountEntries::read_list(void) {

  gchar** device_list = 0;
  gchar** label_list = 0;

  if ((device_list = g_key_file_get_string_list(prog_config.key_file_p,
						"Mounts",
						"Devices",
						0, 0))
      && (label_list = g_key_file_get_string_list(prog_config.key_file_p,
						  "Mounts",
						  "Labels",
						  0, 0))
      && *device_list) {

    gtk_list_store_clear(GTK_LIST_STORE(list_store_h.get()));

    gchar** device_tmp_p;
    gchar** label_tmp_p;
    for (device_tmp_p = device_list, label_tmp_p = label_list;
	 device_tmp_p && *device_tmp_p && label_tmp_p && *label_tmp_p;
	 ++device_tmp_p, ++label_tmp_p) {


      GtkTreeIter iter;
      gtk_list_store_append(GTK_LIST_STORE(list_store_h.get()), &iter);

      gtk_list_store_set(GTK_LIST_STORE(list_store_h.get()), &iter,
			 ModelColumns::device,
			 g_strstrip(*device_tmp_p),
			 -1);
      gtk_list_store_set(GTK_LIST_STORE(list_store_h.get()), &iter,
			 ModelColumns::label,
			 g_strstrip(*label_tmp_p),
			 -1);
    }
  }
  // nothing in this function can throw - this is fine
  // g_strfreev() does nothing if the argument is NULL
  g_strfreev(device_list);
  g_strfreev(label_list);
}

void MountEntries::save_list(void) {

  std::ofstream fileout(prog_config.rc_filename.c_str(), std::ios::out);

  if (fileout) {

    // use std::vector, as the standard requires it to have
    // contiguous memory so it can be passed as an array argument
    std::vector<gchar*> device_list;
    std::vector<gchar*> label_list;

    gchar* device_p;
    gchar* label_p;
    GtkTreeIter iter;
    bool not_at_end = gtk_tree_model_get_iter_first(list_store_h, &iter);
    try {
      while(not_at_end) {
	gtk_tree_model_get(list_store_h, &iter,
			   ModelColumns::device, &device_p,
			   ModelColumns::label, &label_p,
			   -1);
	device_list.push_back(device_p);
	label_list.push_back(label_p);
	not_at_end = gtk_tree_model_iter_next(list_store_h, &iter);
      }
      if (!device_list.empty()) {
	g_key_file_set_string_list(prog_config.key_file_p,
				   "Mounts",
				   "Devices",
				   &device_list[0],
				   device_list.size());
	g_key_file_set_string_list(prog_config.key_file_p,
				   "Mounts",
				   "Labels",
				   &label_list[0],
				   label_list.size());
      }
      else {
	g_key_file_remove_key(prog_config.key_file_p,
			      "Mounts",
			      "Devices",
			      0);
	g_key_file_remove_key(prog_config.key_file_p,
			      "Mounts",
			      "Labels",
			      0);
      }
      GcharScopedHandle data_h(g_key_file_to_data(prog_config.key_file_p, 0, 0));
      fileout << (char*)data_h.get();
    }
    catch (...) {
      // this is just to make sure we clean up
      // the strings in the two vector objects
      // in the next two 'for' statements
      // we shouldn't actually ever be here so it is
      // a bit of a formalism
    }
    for (std::vector<gchar*>::const_iterator iter = device_list.begin();
	 iter != device_list.end();
	 ++iter) {
      g_free(*iter);
    }
    for (std::vector<gchar*>::const_iterator iter = label_list.begin();
	 iter != label_list.end();
	 ++iter) {
      g_free(*iter);
    }
  }
  else {
    std::string message("Can't open file ");
    message += prog_config.rc_filename;
    message += " for saving mount entries\n";
    write_error(message.c_str(), true);
  }    
}

void MountEntries::move_up_impl(void) {

  GtkTreeIter selected_iter;
  GtkTreeModel* model_p;
  GtkTreeSelection* selection_p = gtk_tree_view_get_selection(tree_view_p);

  if (gtk_tree_selection_get_selected(selection_p, &model_p, &selected_iter)) {
    TreePathScopedHandle path_h(gtk_tree_model_get_path(model_p, &selected_iter));
    if (gtk_tree_path_prev(path_h)) {

      GtkTreeIter prev_iter;
      
      if (!gtk_tree_model_get_iter(model_p, &prev_iter, path_h)) {
	write_error("Iterator error in MountEntries::move_up_impl()\n", false);
	beep();
      }
      else {
	gchar* selected_device_p = 0;
	gchar* selected_label_p = 0;
	gchar* prev_device_p = 0;
	gchar* prev_label_p = 0;

	// just do a swap of values
	gtk_tree_model_get(model_p, &selected_iter,
			   ModelColumns::device, &selected_device_p,
			   ModelColumns::label, &selected_label_p,
			   -1);

	gtk_tree_model_get(model_p, &prev_iter,
			   ModelColumns::device, &prev_device_p,
			   ModelColumns::label, &prev_label_p,
			   -1);

	gtk_list_store_set(GTK_LIST_STORE(model_p), &selected_iter,
			   ModelColumns::device, prev_device_p,
			   ModelColumns::label, prev_label_p,
			   -1);
      
	gtk_list_store_set(GTK_LIST_STORE(model_p), &prev_iter,
			   ModelColumns::device, selected_device_p,
			   ModelColumns::label, selected_label_p,
			   -1);

	// we have not placed the gchar* strings given by gtk_tree_model_get()
	// in a GcharScopedHandle object so we need to free them by hand
	g_free(selected_device_p);
	g_free(selected_label_p);
	g_free(prev_device_p);
	g_free(prev_label_p);

	gtk_tree_selection_select_iter(selection_p, &prev_iter);
	changed = true;
      }
    }
    else beep();
  }
  else beep();
}

void MountEntries::move_down_impl(void) {

  GtkTreeIter selected_iter;
  GtkTreeModel* model_p;
  GtkTreeSelection* selection_p = gtk_tree_view_get_selection(tree_view_p);

  if (gtk_tree_selection_get_selected(selection_p, &model_p, &selected_iter)) {
    GtkTreeIter next_iter = selected_iter;;
    if (gtk_tree_model_iter_next(model_p, &next_iter)) {
      
      gchar* selected_device_p = 0;
      gchar* selected_label_p = 0;
      gchar* next_device_p = 0;
      gchar* next_label_p = 0;

      // just do a swap of values
      gtk_tree_model_get(model_p, &selected_iter,
			 ModelColumns::device, &selected_device_p,
			 ModelColumns::label, &selected_label_p,
			 -1);

      gtk_tree_model_get(model_p, &next_iter,
			 ModelColumns::device, &next_device_p,
			 ModelColumns::label, &next_label_p,
			 -1);

      gtk_list_store_set(GTK_LIST_STORE(model_p), &selected_iter,
			 ModelColumns::device, next_device_p,
			 ModelColumns::label, next_label_p,
			 -1);
      
      gtk_list_store_set(GTK_LIST_STORE(model_p), &next_iter,
			 ModelColumns::device, selected_device_p,
			 ModelColumns::label, selected_label_p,
			 -1);

      // we have not placed the gchar* strings given by gtk_tree_model_get()
      // in a GcharScopedHandle object so we need to free them by hand
      g_free(selected_device_p);
      g_free(selected_label_p);
      g_free(next_device_p);
      g_free(next_label_p);

      gtk_tree_selection_select_iter(selection_p, &next_iter);
      changed = true;
    }
    else beep();
  }
  else beep();
}

void TableEntryCB::new_entry_selected(GtkWidget* widget_p, void* data) {
  TableEntry* instance_p = static_cast<TableEntry*>(data);

  if (widget_p == instance_p->ok_button_p) {
    if (instance_p->selected_impl()) {
      instance_p->accepted = true;
      instance_p->close();
    }
  }
  else if (widget_p == instance_p->cancel_button_p) instance_p->close();
  else {
    write_error("Callback error in TableEntryCB::new_entry_selected()\n", false);
    instance_p->close();
  }
}


gboolean TableEntryCB::new_entry_key_press_event(GtkWidget*, GdkEventKey* event_p, void* data) {
  TableEntry* instance_p = static_cast<TableEntry*>(data);
  int keycode = event_p->keyval;
#if GTK_CHECK_VERSION(2,20,0)
  bool cancel_focus = gtk_widget_has_focus(instance_p->cancel_button_p);
#else
  bool cancel_focus = GTK_WIDGET_HAS_FOCUS(instance_p->cancel_button_p);
#endif

#if GTK_CHECK_VERSION(2,99,0)
  if (keycode == GDK_KEY_Return && !cancel_focus) {
    if (instance_p->selected_impl()) {
      instance_p->accepted = true;
      instance_p->close();
    }
    return true;  // stop processing here
  }
#else
  if (keycode == GDK_Return && !cancel_focus) {
    if (instance_p->selected_impl()) {
      instance_p->accepted = true;
      instance_p->close();
    }
    return true;  // stop processing here
  }
#endif
  return false;   // continue processing key events
}

TableEntry::TableEntry(GtkWindow* parent_p,  const gchar* device_p, const gchar* label_p):
      WinBase((device_p && label_p) ? gettext("Mount-gtk: modify entry")
	                            : gettext("Mount-gtk: new entry"),
	      prog_config.window_icon_h,
	      true, parent_p),
      accepted(false) {
                             
  GtkTable* table_p = GTK_TABLE(gtk_table_new(3, 2, false));
  gtk_container_add(GTK_CONTAINER(get_win()), GTK_WIDGET(table_p));


  GtkWidget* device_label_p = gtk_label_new(gettext("Device:"));
  gtk_table_attach(table_p, device_label_p,
		   0, 1, 0, 1,
		   GTK_FILL, GTK_SHRINK,
		   standard_size/2, standard_size/4);
  device_entry_p = gtk_entry_new();
  gtk_table_attach(table_p, device_entry_p,
		   1, 2, 0, 1,
		   GtkAttachOptions(GTK_FILL | GTK_EXPAND), GTK_SHRINK,
		   standard_size/2, standard_size/4);
  GtkWidget* label_label_p = gtk_label_new(gettext("Label:"));
  gtk_table_attach(table_p, label_label_p,
		   0, 1, 1, 2,
		   GTK_FILL, GTK_SHRINK,
		   standard_size/2, standard_size/4);
  label_entry_p = gtk_entry_new();
  gtk_table_attach(table_p, label_entry_p,
		   1, 2, 1, 2,
		   GtkAttachOptions(GTK_FILL | GTK_EXPAND), GTK_SHRINK,
		   standard_size/2, standard_size/4);
  GtkWidget* button_box_p = gtk_hbutton_box_new();
  gtk_table_attach(table_p, button_box_p,
		   0, 2, 2, 3,
		   GtkAttachOptions(GTK_FILL | GTK_EXPAND), GTK_SHRINK,
		   standard_size/2, standard_size/4);
  
  gtk_label_set_justify(GTK_LABEL(device_label_p), GTK_JUSTIFY_RIGHT);
  gtk_label_set_justify(GTK_LABEL(label_label_p), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment(GTK_MISC(device_label_p), 1, 0.5);
  gtk_misc_set_alignment(GTK_MISC(label_label_p), 1, 0.5);
  gtk_widget_set_size_request(device_entry_p, standard_size * 8, -1);
  gtk_widget_set_size_request(label_entry_p, standard_size * 8, -1);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box_p), GTK_BUTTONBOX_END);
  gtk_box_set_spacing(GTK_BOX(button_box_p), standard_size/2);

  if (device_p && label_p) {
    gtk_entry_set_text(GTK_ENTRY(device_entry_p), device_p);
    gtk_entry_set_text(GTK_ENTRY(label_entry_p), label_p);
  }

  cancel_button_p = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_container_add(GTK_CONTAINER(button_box_p), cancel_button_p);
  ok_button_p = gtk_button_new_from_stock(GTK_STOCK_OK);
  gtk_container_add(GTK_CONTAINER(button_box_p), ok_button_p);

  g_signal_connect(G_OBJECT(ok_button_p), "clicked",
		   G_CALLBACK(TableEntryCB::new_entry_selected), this);
  g_signal_connect(G_OBJECT(cancel_button_p), "clicked",
		   G_CALLBACK(TableEntryCB::new_entry_selected), this);
  g_signal_connect(G_OBJECT(get_win()), "key_press_event",
		   G_CALLBACK(TableEntryCB::new_entry_key_press_event), this);

#if GTK_CHECK_VERSION(2,20,0)
  gtk_widget_set_can_default(ok_button_p, true);
  gtk_widget_set_can_default(cancel_button_p, true);
#else
  GTK_WIDGET_SET_FLAGS(ok_button_p, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS(cancel_button_p, GTK_CAN_DEFAULT);
#endif

  gtk_window_set_type_hint(get_win(), GDK_WINDOW_TYPE_HINT_DIALOG);

  gtk_container_set_border_width(GTK_CONTAINER(get_win()), standard_size/2);

  gtk_widget_grab_focus(device_entry_p);

  gtk_window_set_position(get_win(), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_window_set_resizable(get_win(), false);

  gtk_widget_show_all(GTK_WIDGET(get_win()));
}

int TableEntry::get_exec_val(void) const {
  return accepted;
}

bool TableEntry::selected_impl(void) {

  bool return_val = false;

  const gchar* device_text = gtk_entry_get_text(GTK_ENTRY(device_entry_p));
  const gchar* label_text = gtk_entry_get_text(GTK_ENTRY(label_entry_p));

  if (!*device_text || !*label_text) beep();
  
  else {
    result.first = (const char*)device_text;
    result.second = (const char*)label_text;
    strip(result.first);
    strip(result.second);
    return_val = true;
  }
  return return_val;
}
