/*
  Copyright (C) 2009, 2010, 2012 and 2015 Chris Vine

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <string.h>

#include <memory>

#include "fstab_parser.h"
#include "mounter.h"

#include <c++-gtk-utils/pipes.h>
#include <c++-gtk-utils/fdstream.h>
#include <c++-gtk-utils/thread.h>
#include <c++-gtk-utils/shared_handle.h>
#include <c++-gtk-utils/cgu_config.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif


void MounterCB::mounter_mounts_changed(GUnixMountMonitor*, void* data) {
  Mounter* instance_p = static_cast<Mounter*>(data);
  instance_p->fill_cache();
  instance_p->monitor_cb();
}

Mounter::Mounter(void): monitor_h(g_unix_mount_monitor_new()) {

#if GLIB_CHECK_VERSION(2,18,0)
  g_unix_mount_monitor_set_rate_limit(monitor_h, 500);
#endif
  g_signal_connect(G_OBJECT(monitor_h.get()), "mounts_changed",
		   G_CALLBACK(MounterCB::mounter_mounts_changed), this);
  fill_cache();
}

void Mounter::fill_cache(void) {

  mounted_device_set.clear();

  GList* mount_list_p = g_list_first(g_unix_mounts_get(0));
  GList* temp_p = mount_list_p;
  while (temp_p) {
    mounted_device_set.insert((char*)g_unix_mount_get_device_path(static_cast<GUnixMountEntry*>(temp_p->data)));
    g_unix_mount_free(static_cast<GUnixMountEntry*>(temp_p->data));
    temp_p = g_list_next(temp_p);
  }
  g_list_free(mount_list_p);
}

// 'mount' can block for a considerable period when mounting a network
// file system in an attempt to cater for network perturbation.
// Therefore, to avoid blocking the GTK+ user interface,
// Mounter::mount() starts a new thread to execute
// Mounter::mount_thread() which will exec() 'pmount' or 'mount'.  We
// observe the POSIX requirement for multi-threaded programs that the
// only functions which may be called in the child process between the
// fork() and the exec() are async-signal-safe ones.  We communicate
// with the UI thread with Callback::post().
void Mounter::mount(const std::string& device, const std::string& label,
		    Callback::Callback* on_fail) {

  // if c++-gtk-utils-2.0 is being used, we know we are
  // compiling under C++0x/C++11
#if CGU_API_VERSION>=20
  std::unique_ptr<Callback::Callback> cb_a(on_fail);
#else
  std::auto_ptr<Callback::Callback> cb_a(on_fail);
#endif

  if (is_mounted(device)) {
    cb_a->dispatch();
    return;
  }

  // now block off the signals with which the main thread may be interested
  // rather than this one
  sigset_t sig_mask;
  sigemptyset(&sig_mask);
  sigaddset(&sig_mask, SIGCHLD);
  sigaddset(&sig_mask, SIGQUIT);
  sigaddset(&sig_mask, SIGTERM);
  sigaddset(&sig_mask, SIGINT);
  sigaddset(&sig_mask, SIGHUP);
  pthread_sigmask(SIG_BLOCK, &sig_mask, 0);

  char* device_arg = strdup(device.c_str());
  char* label_arg =  strdup(label.c_str());

  if (!Thread::Thread::start(Callback::make(&Mounter::mount_thread,
					    std::pair<char*, char*>(device_arg, label_arg),
					    cb_a.release()),
			     false).get()) {
    free(device_arg);
    free(label_arg);
    if (cb_a.get()) Callback::post(cb_a.release());
    write_error("Cannot start new thread to mount device\n", false);
  }

  // now unblock the signals so that the initial (GUI) thread can receive them
  pthread_sigmask(SIG_UNBLOCK, &sig_mask, 0);
}

void Mounter::mount_thread(std::pair<char*, char*> arg_pair,
			   Callback::Callback* on_fail) {

  ScopedHandle<char*, CFree> device_h(arg_pair.first);
  ScopedHandle<char*, CFree> label_h(arg_pair.second);
  ScopedHandle<char*, CFree> mount_point_h;
  // if c++-gtk-utils-2.0 is being used, we know we are
  // compiling under C++0x/C++11
#if CGU_API_VERSION>=20
  std::unique_ptr<Callback::Callback> cb_a(on_fail);
#else
  std::auto_ptr<Callback::Callback> cb_a(on_fail);
#endif

  // if the device to be mounted is not a block device, we need to
  // mount the device via its mount point specified in /etc/fstab
  // rather than by its device name, or some versions of mount will
  // complain with FUSE file systems for a non-root mounter
  struct stat statbuf;
  if (stat(device_h.get(), &statbuf) || !S_ISBLK(statbuf.st_mode)) {
    mount_point_h.reset(get_fstab_mount_point_for_device(device_h.get()));
    if (!mount_point_h.get()) {
      if (cb_a.get()) Callback::post(cb_a.release());
      return; // get_fstab_mount_point_for_device() will already have
              // displayed an error message
    }
  }

  PipeFifo fork_pipe(PipeFifo::block);

  // now fork to create the process which invokes 'pmount' or 'mount'
  pid_t pid = fork();

  if (pid == -1) {
    if (cb_a.get()) Callback::post(cb_a.release());
    Callback::post(Callback::make(write_error,
				  "Fork error in Mounter::mount_thread(\n",
				  false));
    return;
  }
  
  if (!pid) { // child process
  
    // unblock signals for pmount
    sigset_t sig_mask;
    sigemptyset(&sig_mask);
    sigaddset(&sig_mask, SIGCHLD);
    sigaddset(&sig_mask, SIGQUIT);
    sigaddset(&sig_mask, SIGTERM);
    sigaddset(&sig_mask, SIGINT);
    sigaddset(&sig_mask, SIGHUP);
    // this child process is single threaded, so we can use sigprocmask()
    // rather than pthread_sigmask() (and should do so as sigprocmask()
    // is guaranteed to be async-signal-safe)
    // this process will not be receiving interrupts so we do not need
    // to test for EINTR on the call to sigprocmask()
    sigprocmask(SIG_UNBLOCK, &sig_mask, 0);

    fork_pipe.connect_to_stderr();
    
    if (!mount_point_h.get())
      execlp("pmount", "pmount", device_h.get(), label_h.get(), static_cast<char*>(0));
    else
      execlp("mount", "mount", mount_point_h.get(), static_cast<char*>(0));

    // if we reached this point, then the execlp() call must have failed
    const char message[] = "Exec failed in Mounter::mount_thread().\n"
                           "Is pmount installed?\n";
    write(2, message, sizeof(message) - 1);
    _exit(EXIT_SUCCESS - 1);
  }

  // read from the pipe
  fork_pipe.make_readonly();

  fdistream filein(fork_pipe.get_read_fd(), false); // fork_pipe will manage the file
                                                    // descriptor, not filein
  std::string line;
  std::string err;
  // if we use std::getline() we will be safe against
  // multi-byte encodings such as UTF-8
  while (std::getline(filein, line)) {
    err += line;
    err += '\n';
  }

  pid_t ret;
  int status;
  do {
    ret = waitpid(pid, &status, 0);
  } while (ret == -1 && errno == EINTR);
  
  bool success = false;
  if (ret != -1                                    // waitpid() did not return an error
      && WIFEXITED(status)                         // pmount exited normally
      && WEXITSTATUS(status) == EXIT_SUCCESS) {    // pmount returned success
    success = true;
  }
  if (!success) {
    if (!err.empty()) err += '\n';
    err += gettext("Error mounting following device: ");
    err += device_h.get();
    err += '\n';
    if (cb_a.get()) Callback::post(cb_a.release());
    Callback::post(Callback::make(write_error_del,
				  const_cast<const char*>(strdup(err.c_str())),
				  true));
  }
}

// 'umount' can block for a considerable period when mounting a
// network file system in an attempt to cater for network
// perturbation.  Therefore, to avoid blocking the GTK+ user
// interface, Mounter::unmount() starts a new thread to execute
// Mounter::unmount_thread() which will exec() 'pumount' or 'umount'.
// We observe the POSIX requirement for multi-threaded programs that
// the only functions which may be called in the child process between
// the fork() and the exec() are async-signal-safe ones.  We
// communicate with the UI thread with Callback::post().
void Mounter::unmount(const std::string& device, Callback::Callback* on_fail) {

  // if c++-gtk-utils-2.0 is being used, we know we are
  // compiling under C++0x/C++11
#if CGU_API_VERSION>=20
  std::unique_ptr<Callback::Callback> cb_a(on_fail);
#else
  std::auto_ptr<Callback::Callback> cb_a(on_fail);
#endif

  if (!is_mounted(device)) {
    cb_a->dispatch();
    return;
  }

  // now block off the signals with which the main thread may be interested
  // rather than this one
  sigset_t sig_mask;
  sigemptyset(&sig_mask);
  sigaddset(&sig_mask, SIGCHLD);
  sigaddset(&sig_mask, SIGQUIT);
  sigaddset(&sig_mask, SIGTERM);
  sigaddset(&sig_mask, SIGINT);
  sigaddset(&sig_mask, SIGHUP);
  pthread_sigmask(SIG_BLOCK, &sig_mask, 0);

  char* device_arg = strdup(device.c_str());

  if (!Thread::Thread::start(Callback::make(&Mounter::unmount_thread,
					    device_arg, cb_a.release()),
			     false).get()) {
    free(device_arg);
    if (cb_a.get()) Callback::post(cb_a.release());
    write_error("Cannot start new thread to unmount device\n", false);
  }

  // now unblock the signals so that the initial (GUI) thread can receive them
  pthread_sigmask(SIG_UNBLOCK, &sig_mask, 0);
}

void Mounter::unmount_thread(char* device_arg, Callback::Callback* on_fail) {

  ScopedHandle<char*, CFree> device_h(device_arg);
  ScopedHandle<char*, CFree> mount_point_h;
  // if c++-gtk-utils-2.0 is being used, we know we are
  // compiling under C++0x/C++11
#if CGU_API_VERSION>=20
  std::unique_ptr<Callback::Callback> cb_a(on_fail);
#else
  std::auto_ptr<Callback::Callback> cb_a(on_fail);
#endif

  struct stat statbuf;
  if (stat(device_h.get(), &statbuf) || !S_ISBLK(statbuf.st_mode)) {
    mount_point_h.reset(get_fstab_mount_point_for_device(device_h.get()));
    if (!mount_point_h.get()) {
      if (cb_a.get()) Callback::post(cb_a.release());
      return; // get_fstab_mount_point_for_device() will already have
              // displayed an error message
    }
  }

  PipeFifo fork_pipe(PipeFifo::block);

  // now fork to create the process which invokes 'pumount' or 'umount'
  pid_t pid = fork();

  if (pid == -1) {
    if (cb_a.get()) Callback::post(cb_a.release());
    Callback::post(Callback::make(write_error,
				  "Fork error in Mounter::unmount_thread(\n",
				  false));
    return;
  }
  
  if (!pid) { // child process
  
    // unblock signals for pmount
    sigset_t sig_mask;
    sigemptyset(&sig_mask);
    sigaddset(&sig_mask, SIGCHLD);
    sigaddset(&sig_mask, SIGQUIT);
    sigaddset(&sig_mask, SIGTERM);
    sigaddset(&sig_mask, SIGINT);
    sigaddset(&sig_mask, SIGHUP);
    // this child process is single threaded, so we can use sigprocmask()
    // rather than pthread_sigmask() (and should do so as sigprocmask()
    // is guaranteed to be async-signal-safe)
    // this process will not be receiving interrupts so we do not need
    // to test for EINTR on the call to sigprocmask()
    sigprocmask(SIG_UNBLOCK, &sig_mask, 0);

    fork_pipe.connect_to_stderr();

    if (!mount_point_h.get())
      execlp("pumount", "pumount", device_h.get(), static_cast<char*>(0));
    else
      execlp("umount", "umount", mount_point_h.get(), static_cast<char*>(0));

    // if we reached this point, then the execlp() call must have failed
    const char message[] = "Exec failed in Mounter::unmount_thread().\n"
                           "Is pumount installed?\n";
    write(2, message, sizeof(message) - 1);
    _exit(EXIT_SUCCESS - 1);
  }

  // read from the pipe
  fork_pipe.make_readonly();

  fdistream filein(fork_pipe.get_read_fd(), false); // fork_pipe will manage the file
                                                    // descriptor, not filein
  std::string line;
  std::string err;
  // if we use std::getline() we will be safe against
  // multi-byte encodings such as UTF-8
  while (std::getline(filein, line)) {
    err += line;
    err += '\n';
  }

  pid_t ret;
  int status;
  do {
    ret = waitpid(pid, &status, 0);
  } while (ret == -1 && errno == EINTR);
  
  bool success = false;
  if (ret != -1                                    // waitpid() did not return an error
      && WIFEXITED(status)                         // pumount exited normally
      && WEXITSTATUS(status) == EXIT_SUCCESS) {    // pumount returned success
    success = true;
  }
  if (!success) {
    if (!err.empty()) err += '\n';
    err += gettext("Error unmounting following device: ");
    err += device_h.get();
    err += '\n';
    if (cb_a.get()) Callback::post(cb_a.release());
  }
  if (!err.empty()) Callback::post(Callback::make(write_error_del,
						  const_cast<const char*>(strdup(err.c_str())),
						  true));
}
